"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var THREE = require("three");
var random_1 = require("./utils/random");
var three_1 = require("three");
var plane_1 = require("./shapes/plane");
var typeCheck_1 = require("./utils/typeCheck");
var ParticleSystem = /** @class */ (function () {
    function ParticleSystem(target, options) {
        this.color = 0xedaa67;
        this.duration = 2000; // in MS
        this.elapsedTime = 0;
        this.initialRotationRange = [
            // one of tuple of vec3<float> or vec3<float>. Values in radians
            new THREE.Vector3(0, 0, 0),
            new THREE.Vector3(Math.PI * 2, Math.PI * 2, Math.PI * 2),
        ];
        this.isPlaying = true;
        this.loop = true;
        this.maxParticleSize = 0.1;
        this.maxParticles = 100;
        this.minParticleSize = 0.1;
        this.particleLifetime = 2000; // in MS
        this.particleQueue = [];
        this.particleVelocity = 1; // units per second
        this.particlesPerSecond = 50;
        this.playOnLoad = true;
        this.radius = new THREE.Vector3(1, 1, 1);
        this.rotationRate = 0; // in radians
        this.scene = null;
        this.shape = new plane_1.default();
        this.worldSpace = true;
        // User Defined Values
        this.color = options.color || this.color;
        this.duration = options.duration || this.duration;
        this.initialRotationRange = options.initialRotationRange || this.initialRotationRange;
        this.isPlaying = typeCheck_1.isBool(options.playOnLoad) ? options.playOnLoad || false : this.isPlaying;
        this.loop = typeCheck_1.isBool(options.loop) ? options.playOnLoad || false : this.loop;
        this.maxParticleSize = options.maxParticleSize || options.minParticleSize || 0.1;
        this.maxParticles = options.maxParticles || this.maxParticles;
        this.minParticleSize = options.minParticleSize || options.maxParticleSize || this.minParticleSize;
        this.particleLifetime = options.particleLifetime || this.particleLifetime;
        this.particleVelocity = options.particleVelocity || this.particleVelocity;
        this.particlesPerSecond = options.particlesPerSecond || this.particlesPerSecond;
        this.radius = options.radius || this.radius;
        this.rotationRate = options.rotationRate || this.rotationRate;
        this.shape = options.shape || this.shape;
        this.target = target;
        this.worldSpace = typeCheck_1.isBool(options.worldSpace) ? options.worldSpace || false : this.worldSpace;
        if (this.worldSpace)
            this.findSceneObject();
        // Member Variables
        this.startTime = Date.now();
    }
    ParticleSystem.prototype.createPaticle = function () {
        var size = random_1.randomBoundedInt(this.minParticleSize, this.maxParticleSize);
        var geometry = new THREE.BoxGeometry(size, size, size);
        var material = new THREE.MeshBasicMaterial({ color: this.color });
        var newParticle = new THREE.Mesh(geometry, material);
        var _a = this.shape.getVertex(), u = _a[0], v = _a[1];
        // todo: get and set global rotation
        newParticle.rotation.x = random_1.randomBoundedFloat(this.initialRotationRange[0].x, this.initialRotationRange[1].x);
        newParticle.rotation.y = random_1.randomBoundedFloat(this.initialRotationRange[0].y, this.initialRotationRange[1].y);
        newParticle.rotation.z = random_1.randomBoundedFloat(this.initialRotationRange[0].z, this.initialRotationRange[1].z);
        if (this.worldSpace && this.scene) {
            var pos = new three_1.Vector3();
            this.target.getWorldPosition(pos);
            pos.add(u);
            newParticle.position.set(pos.x, pos.y, pos.z);
            this.scene.add(newParticle);
        }
        else {
            newParticle.position.set(u.x, u.y, u.z);
            this.target.add(newParticle);
        }
        this.particleQueue.push([Date.now(), newParticle]);
    };
    ParticleSystem.prototype.findSceneObject = function () {
        var cur = this.target;
        while (cur.parent)
            cur = cur.parent;
        this.scene = cur;
    };
    ParticleSystem.prototype.removeParticles = function (particles) {
        for (var _i = 0, particles_1 = particles; _i < particles_1.length; _i++) {
            var _a = particles_1[_i], timestamp = _a[0], mesh = _a[1];
            if (this.worldSpace && this.scene)
                this.scene.remove(mesh);
            else
                this.target.remove(mesh);
        }
    };
    ParticleSystem.prototype.update = function (deltaTime /* 50fps */) {
        var _this = this;
        if (deltaTime === void 0) { deltaTime = 0.02; }
        if (!this.isPlaying)
            return;
        if (!this.loop && this.elapsedTime > this.duration && !this.particleQueue.length) {
            // No more particles being produced
            this.stop();
            return;
        }
        // create new particles
        if (this.loop || this.elapsedTime < this.duration) {
            for (var i = 0; i < this.particlesPerSecond * deltaTime; i++) {
                this.createPaticle();
            }
        }
        // cull old particles
        var timeThreshold = Date.now() - this.particleLifetime;
        for (var i = 0; i < this.particleQueue.length; i++) {
            if (this.particleQueue[i][0] < timeThreshold) {
                var removed = this.particleQueue.splice(0, i + 1);
                this.removeParticles(removed);
            }
        }
        // cull excess particles
        var overThreshold = this.particleQueue.length - this.maxParticles;
        if (overThreshold > 0) {
            var removed = this.particleQueue.splice(0, overThreshold);
            this.removeParticles(removed);
        }
        // update current particles
        this.particleQueue.forEach(function (_a) {
            var timestamp = _a[0], mesh = _a[1];
            mesh.getWorldPosition(mesh.position);
            mesh.position.y += _this.particleVelocity * deltaTime;
        });
        this.elapsedTime = Date.now() - this.startTime;
    };
    ParticleSystem.prototype.play = function () {
        this.startTime = Date.now();
        this.elapsedTime = 0;
        this.isPlaying = true;
    };
    ParticleSystem.prototype.pause = function () {
        this.isPlaying = false;
    };
    ParticleSystem.prototype.clear = function () {
        this.removeParticles(this.particleQueue);
        this.particleQueue = [];
    };
    ParticleSystem.prototype.stop = function () {
        this.pause();
        this.clear();
    };
    return ParticleSystem;
}());
exports.default = ParticleSystem;
